
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "img.h"


/**
#
# Exemple d'utilisation de img.c
#
# usage: exemple -i toto.ppm -o resultat.ppm
#
#
**/

void tourne(float angle,float cx,float cy,int x,int y,float *xf,float *yf) {
	*xf=cos(angle)*(x-cx)+sin(angle)*(y-cy)+cx;
	*yf=cos(angle)*(y-cy)-sin(angle)*(x-cx)+cy;
}

void Rotation_directe(float angle,imginfo *IA,imginfo *IB) {
int x,y,z,xx,yy;
float xf,yf;
	CreateImage(IA->XSize,IA->YSize,IA->ZSize,IB);

	// (x,y) : dans l'image A
	// (xf,yf) : dans l'image B

	for(y=0;y<IA->YSize;y++) {
		for(x=0;x<IA->XSize;x++) {
			tourne(angle,IA->XSize/2.0,IA->YSize/2.0,x,y,&xf,&yf);
			xx=(int)(xf+0.5); // on arondi au plus proche!
			yy=(int)(yf+0.5);

			// (xx,yy) dans l'image IB?
			if( xx<0 || xx>=IB->XSize || yy<0 || yy>=IB->YSize ) continue;

			// copie le pixel IB(xx,yy) dans IA(x,y)
			for(z=0;z<IB->ZSize;z++) {
				IB->Data[(yy*IB->XSize+xx)*IB->ZSize+z]=IA->Data[(y*IA->XSize+x)*IA->ZSize+z];
			}
		}
	}

}



void Rotation_indirecte(float angle,imginfo *IA,imginfo *IB) {
int x,y,z,xx,yy;
float xf,yf;
	CreateImage(IA->XSize,IA->YSize,IA->ZSize,IB);

	// (x,y) : dans l'image B
	// (xf,yf) : dans l'image A

	for(y=0;y<IB->YSize;y++) {
		for(x=0;x<IB->XSize;x++) {
			tourne(-angle,IA->XSize/2.0,IA->YSize/2.0,x,y,&xf,&yf);

			// (xf,yf) dans l'image IA?
			if( ImgCheck(xf,yf,IA)==0 ) continue; // en dehors de l'image!

			// copie le pixel IA(xf,yf) dans IB(x,y)
			for(z=0;z<IB->ZSize;z++) {
				IB->Data[(y*IB->XSize+x)*IB->ZSize+z]= InterpoleImg(xf,yf,z,IA);
			}
		}
	}

}


int main(int argc,char *argv[])
{
int i;
char InName[100];	// image a lire
char OutName[100];	// image en sortie
imginfo IA;		// image lue
imginfo IB;		// image creee (rotation directe)
imginfo IC;		// image creee (rotation indirecte)
int k;
char buf[100];

	/// lecture des parametres d'entrée

	InName[0]=0;
	OutName[0]=0;

	for(i=1;i<argc;i++) {
		if( strcmp("-i",argv[i])==0 && i+1<argc ) {
			strcpy(InName,argv[i+1]);
			i++;continue;
		}
		if( strcmp("-o",argv[i])==0 && i+1<argc ) {
			strcpy(OutName,argv[i+1]);
			i++;continue;
		}
	}

	if( InName[0]==0 ) { printf("Image SVP!\n");exit(0); }

	k=LoadImage(InName,&IA);
	if( k ) { printf("Impossible de lire l'image '%s'\n",InName);exit(0); }

	/// info sur l'image
	printf("--- Image %s ---\n",InName);
	printf("largeur=%d,  hauteur=%d,  nb de couleurs = %d\n",IA.XSize,IA.YSize,IA.ZSize);
	printf("----------------\n");

	Rotation_directe(30.0*M_PI/180.0,&IA,&IB);
	Rotation_indirecte(30.0*M_PI/180.0,&IA,&IC);

	// sauvegarde le resultat
	if( OutName[0] ) {
		sprintf(buf,"%s_directe",OutName);
		SaveImage(buf,&IB);
		sprintf(buf,"%s_indirecte",OutName);
		SaveImage(buf,&IC);
	}

	/// libere la memoire
	FreeImage(&IA);
	FreeImage(&IB);
	FreeImage(&IC);

}



